//
// Author:  Carlo Barth
// Date:    2016/10/21
//
// Purpose: A high-quality, artistic render of quantum dots (QDs) near a
//          photonic crystal surface, interacting with a leaky-mode.
//          The leaky mode is represented by a color-coded volume render.
//          The QDs have an emission proportional to the local field energy
//          density, denoted by a glowing aura and random arrows.
//
// Details: The data is based on a finite element method solve of the
//          photonic crystal on glass. The glass substrate is omitted in the
//          render. See the related paper describing most of the experimental
//          and numerical methods and results:
//
//              http://aip.scitation.org/doi/10.1063/1.4995229
//

// Imports
//--------------------------------------------------------------------------
#include "colors.inc"
#include "textures.inc"
#include "glass.inc"
#include "metals.inc"
#include "rand.inc"
#include "shapes.inc"


//==========================================================================
//                  SECTION: Parameters and settings
//==========================================================================

// Global Settings
//--------------------------------------------------------------------------
#version 3.7;
global_settings{ 
    assumed_gamma 1.0 
}

// Global parameters
//--------------------------------------------------------------------------
// 0=off, 1=fast; 2=medium quality; 3=high qual.; 
// 4= medium qual.+ recursion_limit 2
#declare Radiosity_On  = 2;
#declare HDRlighting = true;
#declare Spot_on = true; // whether to use spot lights
#declare Are_light_on = true; // whether to use area light
#declare Are_light_brightness = 0.4; // area light brightness
#declare Camera_Type = 1; // 1 = wide view, 2 = close-up

// Real geometry parameters in nanometer
#declare a_real = 600.; // lattice constant
#declare d_real = 367.; // hole diameter
#declare h_real = 116.; // phc height
#declare h_sup_real = 116.; // superspace height
#declare r_qds_real = 2.8; // radius quantum dots (QD)
#declare r_qds_aura = 15; // maximum radius factor for QD aura
// extent of the field volume 
#declare field_box_dim_real = <346.41016151400004*2, 610., 600.>*0.99999999;
// translation vector for field volume alignment
#declare field_box_trans_real = <-346.41016151400004, -250., -299.99>*0.99999999;

// Derived parameters and abbreviating expressions
#declare a = a_real/a_real;
#declare d = d_real/a_real;
#declare Edge_radius = d/20.; // hole edge rounding radius
#declare h = h_real/a_real;
#declare h_sup = h_sup_real/a_real;
#declare r_qds = r_qds_real/a_real;
#declare field_box_dim = field_box_dim_real/a_real;
#declare field_box_trans = field_box_trans_real/a_real;
#declare sqrt3 = 0.5*sqrt(3.); // for hexagonal lattice
#declare cx = a/sqrt3/4; // standard x coordinate of hexagon
#declare cy = 0.5*a; // standard y coordinate of hexagon
#declare Repeat_unit_cells = true; // whether to repeat the unit cell
#declare rows = 2; // rows of the unit cell repetition pattern
#declare cols = 2; // columns of the unit cell repetition pattern

// Parameters related to artistic depections, rather than real properties
// Quantum dots
#declare Show_QDs = false; // whether to render the quantum dots
#declare QD_brightness = 5.; // brightness of the quantum dots
#declare r_aura = r_qds*r_qds_aura; // radius QD aura max (in diameters)
#declare N_emission_rays = 0;
// Field render
#declare Show_field = false; // whether to render the field or not
#declare Field_brightness = 2.; // brightness of the field render
#declare Field_interpolation = 1; // brightness of the field render
#declare Focal_blur = false; // whether to use focal blur

// Additional lights settings
#declare QD_lights = true;
#declare QD_lights_brightness = .1;

// Quantum dot positions and intensity data from file
// Reading the field distribution file
//--------------------------------------------------------------------------
#declare N_QDs = 50;
#declare QD_coords = array[N_QDs];
#declare QD_intensities = array[N_QDs];
#declare Index = 0;
#fopen MyFile "qd_choice_50.dat" read
#while (defined(MyFile))
    #read (MyFile, X, Y, Z, QD_intensity)
    #declare QD_coords[Index] = <Y/a_real, (Z-h_sup)/a_real, X/a_real>;
    #declare QD_intensities[Index] = QD_intensity;
    #declare Index = Index + 1;
#end

// Camera
//--------------------------------------------------------------------------
camera {
    #switch (Camera_Type)
        // -----------------------------------------
        #case (1)
        orthographic angle 30
        location <0, 2, -3>
        look_at <0, h, 0>
        #break
        // -----------------------------------------
        #case (2)
        angle 30
        location <0, h*4, -cy*3>
        look_at <0, h/2, 0>
        #break
        // -----------------------------------------
        #case (3)
        orthographic angle 50
        location < 0, 3, 0 > 
        look_at < 0, 0, 0 >
        #break
        // -----------------------------------------
        #case (4)
        orthographic angle 50
        location < 0 , 7, 0 > 
        look_at < 0, 0, 0 >
        #break
        // -----------------------------------------
        #case (5)
        orthographic angle 50
        location < 0 , 1.5, 0 > 
        look_at < 0, 0, 0 >
        #break
    #end // end of switch
    
    // Correct aspect for different image formats
    right x*image_width/image_height
    
    #if (Focal_blur)
        aperture 0.025 // [0...N] larger is narrower depth of field (blurrier)
        blur_samples 100 // number of rays per pixel for sampling
        focal_point < 0, h/2, 0 > // point that is in focus <X,Y,Z>
        confidence 0.95 // [0...<1] when to move on while sampling (smaller is less accurate)
        variance 1/200 // [0...1] how precise to calculate (smaller is more accurate)
    #end
}


//==========================================================================
//                  SECTION: Illumination
//==========================================================================

// Radiosity
//--------------------------------------------------------------------------
#if (Radiosity_On > 0)

global_settings {
    #ifndef ( Rad_Quality )
        #declare Rad_Quality = Radiosity_On;
    #end

    //--------- radiosity settings -------------------------------- ///
    // from POV-Ray samples "scene templates/patio-radio.pov

    #switch (Rad_Quality)
    #case (1)
    radiosity {             // --- Settings 1 (fast) ---
        pretrace_start 0.08
        pretrace_end   0.02
        count 50
        error_bound 0.5
        recursion_limit 1
        media on
    }
    #break
    #case (2)
    radiosity {             // --- Settings 2 (medium quality) ---
        pretrace_start 0.08
        pretrace_end   0.01
        count 120
        error_bound 0.25
        recursion_limit 1
        media on
    }
    #break
    #case (3)
    radiosity {             // --- Settings 3 (high quality) ---
        pretrace_start 0.08
        pretrace_end   0.005
        count 400
        error_bound 0.1
        recursion_limit 1
        media on
    }
    #break
    #case (4)
    radiosity {             // --- Settings 4 (medium quality, recursion_limit)
        pretrace_start 0.08
        pretrace_end   0.005
        count 350
        error_bound 0.15
        recursion_limit 2
        media on
    }
    #break
    #case (5)
    radiosity {             // --- Settings 5 (high quality, recursion_limit )
        pretrace_start 0.08
        pretrace_end   0.005
        count 350
        error_bound 0.1
        recursion_limit 2
        media on
    }
    #break
    #end  // end of switch
    } // end of global settings

    #default{ finish {emission 0. diffuse 0. ambient 0.}}

//--------------------------------------------------------------------------
#else

    #default{ finish{ ambient 0. diffuse 0.9 }
    } // for intel computers

#end

// Light sources
//--------------------------------------------------------------------------

// A group of 3 spot lights
#if (Spot_on)
    light_source{ // greenish
        <1.,15*h,-1> color rgb <.8,1,.8>
        spotlight
        point_at<0.5,0,0>
        radius 60  // hotspot
        tightness 100
        falloff 90
    }

    light_source{ // blueish
        <-1.,15*h,-1> color rgb <.8,.8,1>
        spotlight
        point_at<-0.5,0,0>
        radius 60  // hotspot
        tightness 100
        falloff 90
    }

    light_source{  // "white"
        <0,15*h,-1> color rgb <.8,.8,.8>
        spotlight
        point_at<0.,0,-0.5>
        radius 60  // hotspot
        tightness 100
        falloff 90
    }
#end

#if (Are_light_on)
    // soft light (areal_light) ---------  An area light (creates soft shadows)
    // WARNING: This special light can significantly slow down rendering times!
    light_source { 
        <0,0,0>          
        color rgb <1,1,1>*Are_light_brightness    // light's color 
        area_light            // kind of light source
        <-cx, 0, 0> <cx, 0, 0> // lights spread out across this distance (x * z)
        1, 10                // total number of lights in grid
        adaptive 2          // 0,1,2,3...
        jitter              // adds random softening of light
        //looks_like { object{Tube} }
        translate <0, 1.50, 0>   // <x,y,z> position of light
    } //---------------- end of area_light
#end

//==========================================================================
//                  SECTION: Object description
//==========================================================================

// Textures
//--------------------------------------------------------------------------

// Material of the photonic crystal
#declare t_shiny_metal =
texture {
    pigment { 
        rgb <.55, .42, .41>
    }
    finish { 
        specular 0.2 roughness 0.5
        reflection 0.05
        brilliance 4
    }
}

// Functions
//--------------------------------------------------------------------------

// Spherical exponential decay
#declare Exp_decay = function (x, y, z, Rad_factor) {
    exp( -(pow(x,2) + pow(y,2) + pow(z,2))*60/Rad_factor )
};

// Macros
//--------------------------------------------------------------------------

// The quantum dot macro
#macro QuantumDot(Radius, Radius_aura, Intensity, Origin)
    // The actual quantum dot
    sphere{ 
        <0,0,0>, Radius
        texture{ Glass2 } // end of texture
        translate Origin
    } // end of sphere

    // The glowing aura
    sphere { 
        <0,0,0>,
        Radius_aura*(1+Intensity) //Radius* (Intensity*r_qds_aura + 1)
        pigment { rgbt 1 } hollow
        interior { 
         media {
            emission Intensity*QD_brightness
              density { 
                function { Exp_decay(x, y, z, Radius_aura*(1+Intensity)) }
                color_map {
                    [0.0 rgbt <0,0,0,1>]
                    [0.5 rgbt <0.5, 0.5, 0.7,1>]
                    [1.0 rgbt <1,1,1,1>]
                }
              }
         }
         media {
            absorption 0
         }
        }
        translate Origin
    }
    
    // Additional light source
    #if (QD_lights)
        light_source{ Origin color White*QD_lights_brightness/N_QDs }
    #end
    
//    #end // end of #for loop
#end // ------------------ end of macro

// The quantum dot rays macro
#macro Spiral (ampli, Trans_vec, Rot_vec) // stretches and compresses the spring                       
    #local min_length = r_aura*2 ;
    #local mid_length = ampli + min_length ;
    #local time_test = 0.25 ; //0.25/0.75 shows max/min extention
    #local sprnglngth = mid_length + ampli * sin((clock + time_test) * 2*pi) ;

    object {
        union {
     	#local n_per_rev = 300 ; // spheres per spring revolution
     	#local n_of_rev = 8. ; // total coil count for the spring
     	#local h_per_ref = sprnglngth / n_of_rev ; // rise per revolution
     	#local nr = 0 ; // start loop
     	#while (nr < n_per_rev * n_of_rev) // loop the spring sphere until...
        sphere {
        	< 0, 0, 0 > , 0.5*r_qds * sin(nr/n_per_rev*pi/n_of_rev) // adjusts the sphere diameter
        	translate< r_qds * sin(nr/n_per_rev*pi/n_of_rev) , -nr * h_per_ref / n_per_rev, 0.0 >
            rotate< 0, nr * 360 / n_per_rev, 0 >
            
            // Appearence
            finish { ambient 0. diffuse 1. phong 20 }
            pigment { rgbt <1,1,1,.9> }
            
            /*
            pigment { rgbt <1,1,1,0.1> } hollow
            interior
            { media
                { emission 1 }
            }
            */
		}
	    #local nr = nr + 1 ;
	    #end
	    }
	    rotate Rot_vec
	    translate Trans_vec
	}
#end // ------------------ end of macro

// Loop for triangular lattice
//--------------------------------------------------------------------------
#macro TriangularLoop ( Object, trans_vec )
    #local ii = -cols;
    #while ( ii < cols )
        #local jj = -rows;
        #while ( jj < rows )
            #if ( mod(ii,2) = 0 )
              #local hexshift = 0.;
            #else
              #local hexshift = 0.5*a;
            #end

            object{
              Object
              translate <ii*a*sqrt3, 0., jj*a+hexshift>
              translate trans_vec
            } 

            #local jj = jj+1;
        #end
        #local ii = ii+1;
    #end
#end // ------------------ end of macro

// Spiral rays from the quantum dots
//--------------------------------------------------------------------------
#macro GlowRays (Intensity)
 #declare Stream = seed (123);
 #declare R1 = seed(360);
 union{
  #for (R, 1, N_emission_rays)
   #declare Start = VRand_On_Sphere (Stream)*r_qds*2;
   #declare Rot_rand = <rand(R1)*360, rand(R1)*360, rand(R1)*360>;
   object {Spiral (Intensity, Start, Rot_rand)}
  #end
  no_shadow
 }
#end

// Compound objects
//--------------------------------------------------------------------------

// Computational domain
#declare geo_cd = prism {
    linear_sweep
    linear_spline
    0., // sweep the following shape from here ...
    h+250., // ... up through here
    7, // the number of points making up the shape ...
    <-2*cx,0>, <-cx,cy>, <cx,cy>, <2*cx,0>, <cx,-cy>, <-cx,-cy>, <-2*cx,0>
}

// Silicon membrane for complete PhC definition without unit cell
#declare geo_si_membrane = box { <-100,0,-100>, <100,h,100> }

// Silicon material in the hexagonal unit cell
#declare geo_si = prism {
    linear_sweep
    linear_spline
    0., // sweep the following shape from here ...
    h, // ... up through here
    7, // the number of points making up the shape ...
    <-2*cx,0>, <-cx,cy>, <cx,cy>, <2*cx,0>, <cx,-cy>, <-cx,-cy>, <-2*cx,0>
}

// The two cylinders that make the hole
#declare geo_hole_cylinder = union {
    cylinder{<0, h-Edge_radius, 0>, <0, h+0.0001, 0>, d/2.+Edge_radius}
    cylinder { <0., -0.0001, 0.>, <0., h+0.0001, 0.>, d/2. }    
}

// The torus for the rounded edge of the hole
#declare geo_hole_torus = torus{ 
    d/2.+Edge_radius, Edge_radius 
    translate <0, h-Edge_radius, 0>
}

/*
// Photonic crystal in unit cell by difference of silicon membrane and hole
#declare phc = union{
    difference{
        object{ geo_si }
        cylinder{<0, h-Edge_radius, 0>, <0, h+0.0001, 0>, d/2.+Edge_radius}
        cylinder { <0., -0.0001, 0.>, <0., h+0.0001, 0.>, d/2. }
    } // end of difference
              
    torus{ d/2.+Edge_radius, Edge_radius 
           translate <0, h-Edge_radius, 0> }
} // end of union
*/

// All holes
#declare geo_all_hole_cylinders = union{
    TriangularLoop ( geo_hole_cylinder, <0,0,0> )
}

// All torus
#declare geo_all_torus = union{
    TriangularLoop ( geo_hole_torus, <0,0,0> )
}

#declare phc = union{
    difference{
        object{ geo_si_membrane }
        #if (Repeat_unit_cells)
            object{ geo_all_hole_cylinders }
        #else
            object{ geo_hole_cylinder }
        #end
    }
    #if (Repeat_unit_cells)
        object{ geo_all_torus }
    #else
        object{ geo_hole_torus }
    #end
} // end of union

// The complete superspace volume
#declare geo_superspace = difference {
    box { <-100,0.,-100>, <100,h+h_sup,100> }
    object { phc }
}


// A single unit cell of the photonic crystal
//--------------------------------------------------------------------------
#declare PhC_object = object{
    phc
    texture{ t_shiny_metal }
    normal{ bozo .5
      scale <1,0.25,0.25>*1
      turbulence 2.
    }
}

// Field render in unit cell
//--------------------------------------------------------------------------

#declare Field_object = prism {
    linear_sweep
    linear_spline
    0., // sweep the following shape from here ...
    1., // ... up through here
    7, // the number of points making up the shape ...
    <0.,0.5>, <0.25,1.>, <0.75,1.>, <1.,0.5>, <0.75,0.>, <0.25,0.>, <0.,0.5>
    
    texture {
     finish { diffuse 0 ambient 0 reflection 0 }
     pigment {
       colour rgbft <1.000,1.000,1.00,0.000,1.000>
     }
    }
    
    hollow
    interior{
     ior                 1.000
     caustics            0.000
     dispersion          1.000
     dispersion_samples  7.000
     fade_power          0.000
     fade_distance       0.000
     fade_color          rgb <0.000,0.000,0.000>
     
     // Red
     media {
       method     3
       intervals  10
       samples    1, 1
       confidence 0.900
       variance   0.008
       ratio      0.900
       absorption rgb <0,0,0>
       emission   rgb <1,0,0> * Field_brightness
       aa_threshold 0.050
       aa_level    4
       density {
         density_file df3 "efield_energy_in_superspace_R.df3" 
         interpolate Field_interpolation
       }
     }
     
     // Green
     media {
       method     3
       intervals  10
       samples    1, 1
       confidence 0.900
       variance   0.008
       ratio      0.900
       absorption rgb <0,0,0>
       emission   rgb <0,1,0> * Field_brightness
       aa_threshold 0.050
       aa_level    4
       density {
         density_file df3 "efield_energy_in_superspace_G.df3" 
         interpolate Field_interpolation
       }
     }
     
     // Blue
     media {
       method     3
       intervals  10
       samples    1, 1
       confidence 0.900
       variance   0.008
       ratio      0.900
       absorption rgb <0,0,0>
       emission   rgb <0,0,1> * Field_brightness
       aa_threshold 0.050
       aa_level    4
       density {
         density_file df3 "efield_energy_in_superspace_B.df3" 
         interpolate Field_interpolation
       }
     }
     media {
       absorption 0.
     }
    }
    
    // Set proper size and position
    scale field_box_dim
    translate field_box_trans
}

#declare Field_restricted = intersection {
    object{ geo_cd hollow }
    object{ Field_object }
}

// Union of the QDs
#declare QDs_all = union {   
    #local ii = 0;
    #while ( ii < N_QDs )
        #local Int_i = QD_intensities[ii];
        #local Vec_i = QD_coords[ii];
        QuantumDot( r_qds, r_aura, Int_i, Vec_i)
        object {GlowRays(pow(Int_i,2)*a*0.05) translate Vec_i}
        #local ii = ii+1;
    #end
}

// Union of PhC and field
#declare Field_and_QDs = union {
    #if (Show_field)
        //object{ Field_object }
    #end
    #if (Show_QDs)
        QDs_all
    #end
}

//==========================================================================
//                  SECTION: Objects in scene
//==========================================================================

// HDRI environment
//--------------------------------------------------------------------------

#if (HDRlighting)

    sky_sphere{
        pigment{
            image_map{
                hdr "hdr/Mt-Washington-Hotel-Lookout_Gauss15.hdr"
                gamma 1.
                map_type 1
                interpolate 2
            }
        }
        rotate < 0,20,0> //rotate < 0,-60,0>  // adapt it to your direction
    } // end sky_sphere

    //light_source{ <0, 500, 0> color White*25 }

# else
    // Background
    //background{White*0.6}

    // sky -------------------------------------------------------------------
    sky_sphere{ 
        pigment{ 
            gradient <0,1,0>
            color_map{ [0   color rgb<1,1,1>         ]//White
                       [0.4 color rgb<0.14,0.14,0.56>]//~Navy
                       [0.6 color rgb<0.14,0.14,0.56>]//~Navy
                       [1.0 color rgb<1,1,1>         ]//White
                     }
            scale 2 }
    } // end of sky_sphere
#end

// Floor
//--------------------------------------------------------------------------
plane
{ y-0.001, 0.
  texture{ Polished_Chrome }
  finish { reflection 0.2 ambient 0. }
  pigment { rgb <.05, .05, .1> }
  normal{ bumps 0.03
          scale <1,0.25,0.25>*1
          turbulence 5.6
       }
  
  /*
    material{ 
        texture{ Glass }
        
        /*
        texture { 
            pigment{ rgb<.1,.1,.15>}
            finish {
                reflection .05
            }
        }
        */
    }
   */
}


// Photonic crystal
PhC_object

//  Field render and quantum dots
#if (Repeat_unit_cells)
    union{
        TriangularLoop ( Field_and_QDs, <0,0,0> )
    }
#else
    Field_and_QDs
#end

#if (Show_field)
    object{ Field_object }
#end


